var __create = Object.create;
var __defProp = Object.defineProperty;
var __getOwnPropDesc = Object.getOwnPropertyDescriptor;
var __getOwnPropNames = Object.getOwnPropertyNames;
var __getOwnPropSymbols = Object.getOwnPropertySymbols;
var __getProtoOf = Object.getPrototypeOf;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __propIsEnum = Object.prototype.propertyIsEnumerable;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp.call(b, prop))
      __defNormalProp(a, prop, b[prop]);
  if (__getOwnPropSymbols)
    for (var prop of __getOwnPropSymbols(b)) {
      if (__propIsEnum.call(b, prop))
        __defNormalProp(a, prop, b[prop]);
    }
  return a;
};
var __require = /* @__PURE__ */ ((x) => typeof require !== "undefined" ? require : typeof Proxy !== "undefined" ? new Proxy(x, {
  get: (a, b) => (typeof require !== "undefined" ? require : a)[b]
}) : x)(function(x) {
  if (typeof require !== "undefined")
    return require.apply(this, arguments);
  throw new Error('Dynamic require of "' + x + '" is not supported');
});
var __copyProps = (to, from, except, desc) => {
  if (from && typeof from === "object" || typeof from === "function") {
    for (let key of __getOwnPropNames(from))
      if (!__hasOwnProp.call(to, key) && key !== except)
        __defProp(to, key, { get: () => from[key], enumerable: !(desc = __getOwnPropDesc(from, key)) || desc.enumerable });
  }
  return to;
};
var __toESM = (mod, isNodeMode, target) => (target = mod != null ? __create(__getProtoOf(mod)) : {}, __copyProps(isNodeMode || !mod || !mod.__esModule ? __defProp(target, "default", { value: mod, enumerable: true }) : target, mod));

// src/index.ts
import { EntryPlugin } from "webpack";

// src/util.ts
import { resolve, extname, isAbsolute } from "path";
import { existsSync, readFileSync, utimesSync } from "fs";
import createDebugger from "debug";
import tailwind from "tailwindcss/lib/processTailwindFeatures";
import getModuleDependencies from "tailwindcss/lib/lib/getModuleDependencies";
import resolveConfig from "tailwindcss/resolveConfig";
import normalizePath from "normalize-path";
import fastGlob from "fast-glob";

// src/devtools/server.ts
import http from "http";
import { getPort } from "get-port-please";
import createServer from "tailwind-config-viewer/server";

// src/constants.ts
var PLUGIN_NAME = "tailwindcss-webpack-plugin";
var DEFAULT_DEVTOOLS_PORT = 9999;
var DEFAULT_DEVTOOLS_HOST = "127.0.0.1";
var DEVTOOLS_POST_PATH = "/_tailwindcss_devtools_update";
var DEV_TOOLS_VIRTUAL_ID = "_tailwind-devtools_.js";
var TAILWIND_ENTRY_VIRTUAL_ID = "_tailwind_.css";
var DEFAULT_TAILWIND_ENTRY_CONTENT = `@tailwind base;
@tailwind components;
@tailwind utilities;`;
var DEFAULT_TAILWIND_CONFIG_FILE = "tailwind.config.js";
var COMPILE_ENTRY_PATH = new URL("./templates/_compile-entry_.js", import.meta.url).pathname;
var TAILWIND_ENTRY_PATH = new URL("./templates/_tailwind_.css", import.meta.url).pathname;
var DEVTOOLS_ENTRY_PATH = new URL("./templates/_tailwind-devtools_.js", import.meta.url).pathname;
var DEVTOOLS_LOADER_PATH = new URL("./loaders/devtools-loader.js", import.meta.url).pathname;
var CSS_LOADER_PATH = new URL("./loaders/css-loader.js", import.meta.url).pathname;

// src/devtools/server.ts
var DevtoolsServer = class {
  constructor(options, compiler) {
    this._ensureStart = promiseSingleton(() => new Promise(async (resolve2) => {
      this.port = await getPort(this.port);
      if (!this._listen) {
        this.server.listen(this.port, this.host, () => {
          this._listen = true;
          debug(`[devtools-server]: listen at ${this.host}:${this.port}`);
          process.on("exit", () => {
            this.close();
          });
          resolve2({
            port: this.port,
            host: this.host
          });
        });
      }
    }));
    var _a, _b, _c, _d;
    this.host = (_b = (_a = options.devtools) == null ? void 0 : _a.host) != null ? _b : DEFAULT_DEVTOOLS_HOST;
    this.port = (_d = (_c = options.devtools) == null ? void 0 : _c.port) != null ? _d : DEFAULT_DEVTOOLS_PORT;
    this.server = http.createServer(async (req, res) => {
      const { url, method } = req;
      res.setHeader("Access-Control-Allow-Origin", "*");
      res.setHeader("Access-Control-Allow-Methods", "POST, OPTIONS");
      res.setHeader("Access-Control-Allow-Headers", "Content-Type");
      if (url === DEVTOOLS_POST_PATH) {
        if (method === "POST") {
          try {
            const { type, data: changed } = await parseRequestBody(req);
            if (type === "add-classes" && (changed == null ? void 0 : changed.length)) {
              compiler.$tailwind.service.updateChangedContent([
                {
                  content: `<div class="${changed.join(" ")}"></div>`,
                  extension: "html"
                }
              ]);
              compiler.$tailwind.service.invalidateCssModule();
            }
            res.statusCode = 200;
            res.end();
          } catch (err) {
            debug("[devtools-server] error: ", err);
            res.statusCode = 500;
            res.end(`Internal Server Error ${err}`);
          }
        } else {
          res.statusCode = 200;
          res.end();
        }
      } else {
        const configViewerMiddleware = createServer({
          tailwindConfigProvider: () => compiler.$tailwind.service.tailwindConfig
        }).asMiddleware();
        configViewerMiddleware(req, res);
      }
    });
    this.server.on("error", (err) => {
      debug(`[devtools-server]: listen err`, err);
      throw err;
    });
    this.service = compiler.$tailwind;
    this._listen = false;
  }
  async ensureStart() {
    debug(`[devtools-server]: ensure start server`);
    return this._ensureStart();
  }
  close(callback) {
    if (this._listen) {
      debug(`[devtools-server]: close`);
      this.server.close((err) => callback == null ? void 0 : callback(err));
    } else {
      callback == null ? void 0 : callback();
    }
  }
};

// src/util.ts
var debug = createDebugger(PLUGIN_NAME);
var promiseSingleton = (fn) => {
  let promise;
  return function wrappedPromise() {
    if (!promise) {
      promise = fn();
    }
    return promise;
  };
};
var isString = (v) => typeof v === "string";
var isObject = (v) => Object.prototype.toString.call(v) === "[object Object]";
var loadConfiguration = (options, root = process.cwd()) => {
  let configPath = resolve(root, DEFAULT_TAILWIND_CONFIG_FILE);
  let dependencies = void 0;
  if (options == null ? void 0 : options.config) {
    if (isString(options.config)) {
      configPath = options.config;
    } else {
      return { config: options.config };
    }
  }
  let config = resolveConfig(existsSync(configPath) ? __require(configPath) : {});
  if (existsSync(configPath)) {
    const deps = getModuleDependencies(configPath).map(({ file }) => file);
    dependencies = new Set(deps);
  }
  return {
    config,
    dependencies
  };
};
var getChangedContent = (config) => {
  let changedContent = [];
  const globs = config.content.files.filter((file) => isString(file)).map((glob) => normalizePath(glob));
  changedContent = fastGlob.sync(globs).map((file) => ({
    content: readFileSync(file, "utf8"),
    extension: extname(file).slice(1)
  }));
  const rawContents = config.content.files.filter((file) => isObject(file) && file !== null);
  for (let { raw: content, extension = "html" } of rawContents) {
    changedContent.push({
      content,
      extension
    });
  }
  debug(`[service]: getChangedContent ->`, changedContent.length);
  return changedContent;
};
var createService = (options) => {
  let context = null;
  let configDependencies = /* @__PURE__ */ new Set();
  let tailwindConfig = null;
  let changedContent = [];
  const transformCSS = async (source) => {
    const tailwindPlugin = () => ({
      postcssPlugin: "tailwindcss",
      Once(root, { result }) {
        tailwind(({ createContext }) => {
          return () => {
            if (context !== null) {
              context.changedContent = changedContent;
              return context;
            }
            context = createContext(tailwindConfig, changedContent);
            return context;
          };
        })(root, result);
      }
    });
    tailwindPlugin.postcss = true;
    const processor = (await Promise.resolve().then(() => __toESM(__require("postcss")))).default([tailwindPlugin]);
    return processor.process(source, { from: void 0 });
  };
  const updateChangedContent = (changed) => {
    debug(`[service]: update chaned content: `, changed.length);
    changedContent.push(...changed);
  };
  const clearChangedContent = () => changedContent = [];
  const getCompletions = () => {
    const classes = context.getClassList().map((className) => `.${className}{}`).join(" ");
    return `
        const style = document.createElement('style');
        style.setAttribute('type', 'text/css');
        style.innerHTML = ${JSON.stringify(classes)};
        const head = document.getElementsByTagName('head')[0];
        head.prepend(style);
    `;
  };
  const ensureInit = promiseSingleton(async () => {
    debug(`[service]: ensure init service`);
    const { config, dependencies } = loadConfiguration(options);
    tailwindConfig = config;
    if (dependencies == null ? void 0 : dependencies.size) {
      for (const dependency of dependencies) {
        configDependencies.add(dependency);
      }
    }
    changedContent = getChangedContent(tailwindConfig);
    if (context === null) {
      await transformCSS(DEFAULT_TAILWIND_ENTRY_CONTENT);
    }
  });
  const invalidateCssModule = () => {
    var _a, _b;
    debug(`invalidate virtual module ${(_a = options.entry) != null ? _a : TAILWIND_ENTRY_PATH}`);
    const fullPath = ensureAbsolute((_b = options.entry) != null ? _b : TAILWIND_ENTRY_PATH);
    const timestamp = Date.now();
    utimesSync(fullPath, timestamp, timestamp);
  };
  const refresh = async () => {
    debug("[service]: refresh");
    configDependencies.clear();
    changedContent = [];
    tailwindConfig = null;
    const { config, dependencies } = loadConfiguration(options);
    tailwindConfig = config;
    if (dependencies == null ? void 0 : dependencies.size) {
      for (const dependency of dependencies) {
        configDependencies.add(dependency);
      }
    }
    changedContent = getChangedContent(tailwindConfig);
  };
  return {
    get context() {
      return context;
    },
    get tailwindConfig() {
      return tailwindConfig;
    },
    configDependencies,
    transformCSS,
    updateChangedContent,
    clearChangedContent,
    getCompletions,
    ensureInit,
    invalidateCssModule,
    refresh
  };
};
var ensureService = async (compiler, options) => {
  if (compiler.$tailwind) {
    return;
  }
  debug(`ensure compiler.$tailwind options -> `, options);
  compiler.$tailwind = {
    dirty: /* @__PURE__ */ new Set(),
    server: new DevtoolsServer(options, compiler),
    service: createService(options)
  };
};
var parseRequestBody = (req) => new Promise((resolve2, reject) => {
  let data = "";
  req.on("data", (chunk) => data += chunk);
  req.on("end", () => {
    try {
      const json = JSON.parse(data);
      resolve2(json);
    } catch (err) {
      reject(err);
    }
  });
  req.on("error", (err) => reject(err));
});
var ensureAbsolute = (file, root = process.cwd()) => isAbsolute(file) ? file : resolve(root, file);

// src/index.ts
var TailwindCSSWebpackPlugin = class {
  constructor(options) {
    this.options = options != null ? options : {};
  }
  apply(compiler) {
    var _a, _b, _c;
    compiler.options.resolve.alias = __spreadValues({
      [DEV_TOOLS_VIRTUAL_ID]: DEVTOOLS_ENTRY_PATH,
      [TAILWIND_ENTRY_VIRTUAL_ID]: ensureAbsolute((_b = (_a = this.options) == null ? void 0 : _a.entry) != null ? _b : TAILWIND_ENTRY_PATH)
    }, compiler.options.resolve.alias);
    compiler.options.module.rules.unshift({
      enforce: "pre",
      include: (resource) => resource.includes(TAILWIND_ENTRY_VIRTUAL_ID) || /\.css$/.test(resource) && !/node_modules/.test(resource),
      loader: CSS_LOADER_PATH
    });
    compiler.options.module.rules.unshift({
      enforce: "pre",
      include: (resource) => resource.includes(DEV_TOOLS_VIRTUAL_ID),
      loader: DEVTOOLS_LOADER_PATH
    });
    if (!isString((_c = this.options) == null ? void 0 : _c.entry)) {
      new EntryPlugin(compiler.context, COMPILE_ENTRY_PATH, {
        name: void 0
      }).apply(compiler);
    }
    compiler.hooks.invalid.tap(PLUGIN_NAME, (file) => {
      var _a2, _b2, _c2;
      debug(`[compiler ${(_a2 = compiler.name) != null ? _a2 : ""}] invalid hook file ->`, file);
      if (!file) {
        file = "config-file";
      }
      if (file === ensureAbsolute((_c2 = (_b2 = this.options) == null ? void 0 : _b2.entry) != null ? _c2 : TAILWIND_ENTRY_PATH)) {
        return;
      }
      if (!compiler.$tailwind.dirty.has(file)) {
        compiler.$tailwind.dirty.add(file);
        compiler.$tailwind.service.invalidateCssModule();
      }
    });
    compiler.hooks.afterCompile.tap(PLUGIN_NAME, () => {
      compiler.$tailwind.service.clearChangedContent();
    });
    compiler.hooks.beforeCompile.tapPromise(PLUGIN_NAME, async () => {
      var _a2;
      debug(`[compiler ${(_a2 = compiler.name) != null ? _a2 : ""}] beforeCompiler hook`);
      await ensureService(compiler, this.options);
    });
    compiler.hooks.watchRun.tapPromise(PLUGIN_NAME, async () => {
      var _a2;
      debug(`[compiler ${(_a2 = compiler.name) != null ? _a2 : ""}] watchRun hook`);
      await ensureService(compiler, this.options);
    });
    compiler.hooks.beforeRun.tapPromise(PLUGIN_NAME, async () => {
      var _a2;
      debug(`[compiler ${(_a2 = compiler.name) != null ? _a2 : ""}] beforeRun hook`);
      await ensureService(compiler, this.options);
    });
    compiler.hooks.shutdown.tapAsync(PLUGIN_NAME, (callback) => {
      var _a2;
      debug(`[compiler ${(_a2 = compiler.name) != null ? _a2 : ""}] shutdown hook`);
      compiler.$tailwind.server.close(callback);
    });
  }
};
export {
  TailwindCSSWebpackPlugin
};
